#include "SerialBridge.h"

SerialBridge::SerialBridge(SC16IS750 *expander, byte dtrPort) : AddOnSerialBase(expander, dtrPort), m_server(0) {
  m_doLogging = true;
}

void SerialBridge::Begin(uint tcpPort) {
  m_tcpPort = tcpPort;
  m_server = WiFiServer(m_tcpPort);

  AddOnSerialBase::Begin();

  m_server.begin();
  m_server.setNoDelay(true);
}

void SerialBridge::Handle() {
  if (m_server.hasClient()) {
    m_client.stop();
    m_client = m_server.available();
    m_client.setNoDelay(true);
    Reset();
    delay(250);
   
    String logItem = "SerialBridge: #" + String(m_tcpPort) + " Client connected, IP=";
    logItem += m_client.remoteIP().toString();
    logItem += " Port=";
    logItem += m_client.remotePort();
    Log(logItem); 
  }
 
  if (m_client.available()) {
    while (m_client.available()) {
      m_expander->Write(m_client.read());
    }
    delay(10);
  }
  

  byte len = m_expander->Available();
  if (len > 0) {
    byte buffer[len];
    for (byte b = 0; b < len; b++) {
      buffer[b] = m_expander->Read();
    }

    if (m_client && m_client.connected()) {
      m_client.write(buffer, len);
    }

  }

}





